package controller.model;

import java.util.LinkedList;
import java.util.List;
import java.util.concurrent.BlockingQueue;

import controller.PrototypeFirstTime;

/**
 * Trieda vytvarajuca zaznamy o zakrutach. Bezi pri tom v inom vlakne
 * ako ovladaci modul vozidla.
 * 
 * @author Ivan Valencik
 */
public class NewTurnDetector implements Runnable
{
	private final BlockingQueue<TickInfo> tickInfoPipe;
	
	private static final LinkedList<NewTurn> turns = new LinkedList<NewTurn>();

	// odporucam tento atribut nepouzivat na rozoznavanie, pouzivany je na synchronizaciu
	// a jeho vyznam sa pri zmene jej implementacie moze zmenit
	private static double _focusValue;

	public NewTurnDetector(BlockingQueue<TickInfo> source)
	{
		tickInfoPipe = source;
	}

	public void run()
	{
		try {
			while (true)
			{
				TickInfo tick = tickInfoPipe.take();
				
				inform(tick);
				//System.out.println("Dlzka radu na spracovanie: " + tickInfoPipe.size());
			}
		} catch (InterruptedException e) {
			System.err.println("Error: chyba pri prijmani akcie do modelu trate.");
			e.printStackTrace();
		}
	}
		
	/**
	 * Metoda vrati zoznam rozoznanych zakrut. <br/><br/> 
	 * Pozor: zoznam sa sa neaktualizuje, pre ziskanie jeho novej verzie je
	 * metodu potrebne zavolat znova.
	 * @return Zoznam zakrut
	 */
	@SuppressWarnings("unchecked")
	public static LinkedList<NewTurn> getTurnList()
	{
		LinkedList<NewTurn> tmp;
		synchronized (turns)
		{
			tmp =(LinkedList<NewTurn>) turns.clone();
		}
		return tmp;
	}

	/**
	 * <p>Metoda vrati hodnotu focusu na urceny uhol. Metoda je blokujuca.</p>
	 * <p>Kazde volanie metody vrati hodnotu ASAP, avsak meranie moze byt vykonane
	 * iba kazdu sekundu herneho casu ( [senzor_model].getCurrentLapTime() ).
	 * @param desiredFocus zelane nastavenie focusu z rozsahu <-90, 90>
	 * @throws TooSoonException vynimka je vyhodena ak bola metoda zavolana skorej ako za
	 * jednu sekundu herneho casu
	 */
	private double requestSpecificFocusValue(int desiredFocus) throws TooSoonException
	{
		PrototypeFirstTime.requestFocusChange(desiredFocus);

		synchronized(PrototypeFirstTime.lock)
		{
			try {
				PrototypeFirstTime.lock.wait();
				if (_focusValue == -3)
					throw new TooSoonException();
				else
					return _focusValue;
			} catch (InterruptedException e) {
				throw new RuntimeException(e);
			}
		}
	}
	
	/**
	 * Tato metoda ma byt volana len v ramci ziadania nastavenia noveho focusu.
	 * @param focusValue hodnota stredneho senzora z focusu
	 */
	public static void setFocusValue(double focusValue)
	{
		_focusValue = focusValue;
	}
	
	private LinkedList<Double> steeringRecord = null;
	private double turnStart;
	
	private static final double MAX_DIFF_RANGE = 0.01;
	private static final int SEQUENCE_LENGTH = 5;
		
	/**
	 * Metoda spravi priemer honot postupnosti, ale iba pokial su vsetky prvky
	 * v rozsahu urcenom konstantou MAX_DIFF_RANGE.
	 * @param subList zoznam hodnot zatoceni volantom
	 * @return priemerna hodnota postupnosti alebo -1, ak nesplnala podmienky
	 */
	private double checkValues(List<Double> subList) {
		double min = 1, max = 0;

		for (double d : subList) {
			if ((d > 0 && subList.get(0) < 0) || (d < 0 && subList.get(0) > 0))
				return -1;

			double d1 = Math.abs(d);
			if (d1 < min)
				min = d1;
			if (d1 > max)
				max = d1;
		}
		
		if (max - min < MAX_DIFF_RANGE) {
			double sum = 0;
			for (double d : subList)
				sum += d;
			return sum / subList.size();
		}

		return -1;
	}

	/**
	 * Na zaklade hodnot natoceni volantu v zakrute priradi k zakrute cislo reprezentujuce
	 * usek najprudsieho zatacania v zakrute.
	 * @return hodnota reprezentujuca zatocenie zakruty
	 */
	private double processSteeringRecord() {
		double val = 0;
		double maxVal = 0;

		for (int i = 0; i <= steeringRecord.size() - SEQUENCE_LENGTH; i++)
			if ((val = checkValues(steeringRecord.subList(i, i + SEQUENCE_LENGTH - 1))) != -1)
			{				
				if (Math.abs(val) > Math.abs(maxVal))
					maxVal = val;
			}
		
		return maxVal;
	}

	/**
	 * Metoda rozoznavajuca zakruty na zaklade riadenim vykonanych cinnosti.
	 * @param tick Informacie o tiku
	 * @throws InterruptedException vynimka vzniknuta pri cakani na zmenu focusu
	 */
	public void inform(TickInfo tick)
	{
		if (steeringRecord == null && tick.isTurning()) {
			System.out.println("zaciatok zakruty");
			steeringRecord = new LinkedList<Double>();
			steeringRecord.add(tick.getAction().steering);
			turnStart = tick.getSensorModel().getDistanceFromStartLine();
		}

		if (steeringRecord != null && tick.isTurning())
			steeringRecord.add(tick.getAction().steering);

		if (steeringRecord != null && !tick.isTurning()) {
			System.out.println("koniec zakruty");
			double val = processSteeringRecord();
			steeringRecord = null;
			synchronized (turns)
			{
				turns.add(new NewTurn(turnStart,
						tick.getSensorModel().getDistanceFromStartLine(), val));
				
				System.out.println(turns.getLast().getDirection().toString() + ": " +
						turns.getLast().getPeakValue());
			}
		}
		
		// Ukazkove volanie na zmenu focusu: metodu volam v priemere kazdy druhy krat.
		// Metoda moze byt volana iba raz za sekundu, cize vacsinou metoda vyhodi vynimku.
//		if ((new java.util.Random()).nextDouble() > 0.5)
//			try {
//				System.out.println("ziskana hodnota: " + requestSpecificFocusValue(12));
//			} catch (TooSoonException e) {
//				//System.out.println("priskoro");
//			}
		

		
	}
}

