package controller.model;

import java.util.ArrayDeque;
import java.util.Deque;
import java.util.LinkedList;
import java.util.concurrent.BlockingQueue;

import controller.PrototypeFirstTime;
import controller.model.Turn.Direction;


import champ2010client.SensorModel;

/**
 * Trieda vytvarajuca zaznamy o zakrutach. Bezi pri tom v inom vlakne
 * ako ovladaci modul vozidla.
 * 
 * @author Ivan Valencik
 */
public class TurnDetector implements Runnable
{
	private boolean focusSet=false;

	private final BlockingQueue<TickInfo> tickInfoPipe;
	// odporucam tento atribut nepouzivat na rozoznavanie, pouzivany je na synchronizaciu
	// a jeho vyznam sa pri zmene jej implementacie moze zmenit
	private static double _focusValue;

	double avgTurn = 0, maxTurn = 0, distance = -1, maxTurnPoint = 0, sector = 0;
	double duration = 0, counter = 0;
	private Deque<Double> uhly = new ArrayDeque<Double>(); 
	/**
	 * pocet tikov, ktore su spajane do jedneho sektora
	 */
	private int SECTOR_LENGTH = 15;
	/**
	 * hodnota natocenia volantu od ktorej je zatacanie povazovane za relevantne
	 */
	private double TURN_THRESHOLD = 0.05;
	/**
	 * minimalny pocet sektorov z ktorych sa sklada zakruta
	 */
	private int MIN_TURN_DURATION = 2;

	private static final LinkedList<Turn> turns = new LinkedList<Turn>();

	public TurnDetector(BlockingQueue<TickInfo> source)
	{
		tickInfoPipe = source;
	}

	public void run()
	{
		try {
			while (true)
			{
				TickInfo tick = tickInfoPipe.take();
				
				if(uhly.size()==10 )
					uhly.pollFirst();
				uhly.addLast(tick.getSensorModel().getAngleToTrackAxis());
				inform(tick);
				//System.out.println("Dlzka radu na spracovanie: " + tickInfoPipe.size());
			}
		} catch (InterruptedException e) {
			System.err.println("Error: chyba pri prijmani akcie do modelu trate.");
			e.printStackTrace();
		}
	}
	private boolean somNaKonci(){
		
		double curValue = uhly.pollFirst().doubleValue();
		for(int i=0;i<uhly.size()-2;i++){
			if(!(curValue<uhly.peekFirst().doubleValue()))
				return false;
			curValue = uhly.pollFirst().doubleValue();
		}
		if(!(curValue>uhly.peekFirst().doubleValue()))
			return true;
		else 
			return false;

	}
	/**
	 * Metoda vrati zoznam rozoznanych zakrut. <br/><br/> 
	 * Pozor: zoznam sa sa neaktualizuje, pre ziskanie jeho novej verzie je
	 * metodu potrebne zavolat znova.
	 * @return Zoznam zakrut
	 */
	@SuppressWarnings("unchecked")
	public static LinkedList<Turn> getTurnList()
	{
		LinkedList<Turn> tmp;
		synchronized (turns)
		{
			tmp =(LinkedList<Turn>) turns.clone();
		}
		return tmp;
	}

	/**
	 * Metoda volana pri zacati sledovania novej zakruty.
	 * @param sm senzory
	 */
	private void newTurn(SensorModel sm)
	{
		duration = 0;
		distance = maxTurnPoint = sm.getDistanceFromStartLine();
		avgTurn  = sector;
		System.out.println("start");
	}

	/**
	 * <p>Metoda vrati hodnotu focusu na urceny uhol. Metoda je blokujuca.</p>
	 * <p>Kazde volanie metody vrati hodnotu ASAP, avsak meranie moze byt vykonane
	 * iba kazdu sekundu herneho casu ( [senzor_model].getCurrentLapTime() ).
	 * @param desiredFocus zelane nastavenie focusu z rozsahu <-90, 90>
	 * @throws TooSoonException vynimka je vyhodena ak bola metoda zavolana skorej ako za
	 * jednu sekundu herneho casu
	 */
	private double requestSpecificFocusValue(int desiredFocus) throws TooSoonException
	{
		PrototypeFirstTime.requestFocusChange(desiredFocus);

		synchronized(PrototypeFirstTime.lock)
		{
			try {
				PrototypeFirstTime.lock.wait();
				if (_focusValue == -3)
					throw new TooSoonException();
				else
					return _focusValue;
			} catch (InterruptedException e) {
				throw new RuntimeException(e);
			}
		}
	}
	
	/**
	 * Tato metoda ma byt volana len v ramci ziadania nastavenia noveho focusu.
	 * @param focusValue hodnota stredneho senzora z focusu
	 */
	public static void setFocusValue(double focusValue)
	{
		_focusValue = focusValue;
	}
	
	/**
	 * Metoda rozoznavajuca zakruty na zaklade riadenim vykonanych cinnosti.
	 * @param tick Informacie o tiku
	 * @throws InterruptedException vynimka vzniknuta pri cakani na zmenu focusu
	 */
	public void inform(TickInfo tick)
	{
		Turn turn = null;

		// Ukazkove volanie na zmenu focusu: metodu volam v priemere kazdy druhy krat.
		// Metoda moze byt volana iba raz za sekundu, cize vacsinou metoda vyhodi vynimku.
//		if ((new java.util.Random()).nextDouble() > 0.5)
//			try {
//				System.out.println("ziskana hodnota: " + requestSpecificFocusValue(12));
//			} catch (TooSoonException e) {
//				//System.out.println("priskoro");
//			}
		
//		double lol=0;
//		if(somNaKonci() && !focusSet){
//			lol =tick.getSensorModel().getDistanceFromStartLine();
//			
//			focusSet=true;
//			
//			return;
//		}
//
//		if(focusSet){
//			turn = new Turn(avgTurn, maxTurn, distance, maxTurnPoint,
//					tick.getSensorModel().getDistanceFromStartLine());
//			distance = -1;
//		}

		
		
		
		sector += tick.getAction().steering;
		if (Math.abs(tick.getAction().steering) > maxTurn)
		{
			maxTurn = Math.abs(tick.getAction().steering);
			maxTurnPoint = tick.getSensorModel().getDistanceFromStartLine();
		}

		if (++counter == SECTOR_LENGTH)
		{
			
			sector /= SECTOR_LENGTH;


			// prebieha zatacanie?
			if (Math.abs(sector) > TURN_THRESHOLD)
			{
				// je zaznamenavana zakruta?
				if (distance != -1)
				{
					// suhlasi zatacanie s uz zaznamenavanym? 
					if (((sector > 0) && (avgTurn > 0)) || ((sector <= 0) && (avgTurn <= 0)))
						avgTurn = controller.Util.rollingAverage(sector, avgTurn, duration++);
					else
					{
						if (duration >= MIN_TURN_DURATION)
							turn = new Turn(avgTurn, maxTurn, distance, maxTurnPoint,
									tick.getSensorModel().getDistanceFromStartLine());
						newTurn(tick.getSensorModel());
					}
				}
				else
					newTurn(tick.getSensorModel());
			}
			else
				if (distance != -1)
				{
					if (duration >= MIN_TURN_DURATION)
						turn = new Turn(avgTurn, maxTurn, distance, maxTurnPoint,
								tick.getSensorModel().getDistanceFromStartLine());
					distance = -1;
				}

			if (turn != null)
				synchronized (turns)
				{
					turns.add(turn);
					maxTurn=0;
					if (turns.getLast().getDirection() == Direction.LEFT)
						System.out.print("Vlavo: ");
					else
						System.out.print("Vpravo: ");
					System.out.println(turns.getLast().getAvgTurn() + " " +
							turns.getLast().getStartDistance() + "-" +
							turns.getLast().getMaxTurnPoint() + "-" +
							turns.getLast().getEndDistance() + " max. turn: " +
							turns.getLast().getMaxTurn()
					);
				}

			sector = 0;
			counter = 0;
		}
	}
}

/**
 * Vynimka upozornujca na to, ze metoda, ktora ju vyhodila bola zavolana skorej,
 * ako je dovolene.
 * @author Ivan Valencik
 */
@SuppressWarnings("serial")
class TooSoonException extends Exception {}
